unit MainUnit;

{$M-}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls, StdCtrls, ExtCtrls;

type
  TConfigForm = class(TForm)
    StarLabel: TLabel;
    StarTrack: TTrackBar;
    SpeedLabel: TLabel;
    SpeedTrack: TTrackBar;
    Panel1: TPanel;
    Panel2: TPanel;
    OKButton: TButton;
    CancelButton: TButton;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure OKButtonClick(Sender: TObject);
    procedure CancelButtonClick(Sender: TObject);
    procedure StarTrackChange(Sender: TObject);
    procedure SpeedTrackChange(Sender: TObject);
  protected
    StarHead, SpeedHead: String;
    bProgrammaticChange: Boolean;
    { Private declarations }
  public
    { Public declarations }
  end;

var
  ViewMode: (vmScreensave, vmPreview, vmConfig);
  hDrawWindow: THandle;
  ConfigForm: TConfigForm;

procedure LoadParams;
procedure StartScreensaver;
procedure StopScreensaver;



implementation

{$R *.DFM}

uses
  Math, Randomity,
  Registry,
  GL, GLU, GLD,
  DUtils, OpenGLApertures, DClocks, D3Vectors;

const
  NearClip = 1;
  FarClip = 6*1024 * NearClip;
  MyKey = 'Software\Dare\SuperStarSaver\2.0';

type
  TColor3F = record
    R, G, B: TReal;
  end;

  TOpenGLApertureWithFrame = class(TOpenGLAperture)
  protected
    procedure IdleFrame(Sender: TObject; var Done: Boolean);
  end;

  TStarData = class
  public
    constructor Create; virtual;
  public
    Loc: T3Vector;
    Color: TColor3F;

    function bInFrustum: Boolean;
    procedure DoMove(M, B: TReal); virtual;
    procedure DoDraw; virtual;
  protected
    SideStep: T3Vector;
    oLoc: T3Vector;
    procedure BeNewStar(bAnywhere: Boolean); virtual;
  end;

var
  Aperture: TOpenGLApertureWithFrame;
  MasterClock: TDClock;
  StarData: array of TStarData;
  bNeedsSort: Boolean;
  cX, cY, cDiag, cwX, cwY: TReal;
  SpeedFactor: TReal = 1024;
  StarCount: Integer = 4096;

procedure LoadParams;
var
  iParam: Integer;
  Command, Param: String;
  ReadState: (rsPlain, rsPreviewHandle);
  Reg: TRegistry;

begin
  ViewMode := vmConfig;
  hDrawWindow := INVALID_HANDLE_VALUE;

  Reg := TRegistry.Create(KEY_QUERY_VALUE);
  Reg.RootKey := HKEY_CURRENT_USER;
  if Reg.OpenKey(MyKey, False) then begin
    try
      SpeedFactor := StrToFloat(Reg.ReadString('SpeedFactor'));
    except
    end;
    try
      StarCount := Reg.ReadInteger('StarCount');
    except
    end;
  end;
  Reg.Free;

  ReadState := rsPlain;
  for iParam := 1 to ParamCount do begin
    SplitStr(ParamStr(iParam), ':', Command, Param);
    if Command = '' then Continue;

    case ReadState of
      rsPlain: if (Command[1] = '/') and (Length(Command) >= 2) then begin
        case UpCase(Command[2]) of
          'C': ViewMode := vmConfig;
          'P': begin
            ViewMode := vmPreview;
            ReadState := rsPreviewHandle;
          end;
          'S': ViewMode := vmScreensave;
        end;
      end
      else begin
        //`@
      end;
      rsPreviewHandle: begin
        ReadState := rsPlain;
        try
          hDrawWindow := StrToInt(Command);
        except
          hDrawWindow := INVALID_HANDLE_VALUE;
        end;
      end;
    end;
  end;
end;

procedure ResizeGLScene(Width, Height: Integer; var cwX, cwY: Double);
var
  AspectRoot, iAspectRoot: Double;

begin
  //Reset the current viewport.
  glViewport(0, 0, Width, Height);

  //Reset the projection matrix.
  glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    AspectRoot := SqRt(Width / Height);
    iAspectRoot := 1 / AspectRoot;
    glFrustum(
      +NearClip/2 *  AspectRoot, -NearClip/2 *  AspectRoot,
      -NearClip/2 * iAspectRoot, +NearClip/2 * iAspectRoot,
      NearClip, FarClip
    );
    gldYaw(180);
  glMatrixMode(GL_MODELVIEW);

  cwX := 1/2 *  AspectRoot;
  cwY := 1/2 * iAspectRoot;
end;

procedure StartScreensaver;
const
  TexSz = 64;
  cLo = (TexSz shr 1) - 1;
  cHi = (TexSz shr 1) - 0;

var
  iStar, iTexel, iX, iY: Integer;
  ToWidth, ToHeight: Integer;
  ClientR: TRect;
  hStarTex: LongWord;
  TexWord: array of Word;
  Err: LongWord;

begin
  Aperture := TOpenGLApertureWithFrame.Create(hDrawWindow);
  MasterClock := TDClock.Create;

  glClearColor(0, 0, 0, 0);

  GetClientRect(hDrawWindow, ClientR);
  with ClientR do begin
    ToWidth := Right - Left;
    ToHeight := Bottom - Top;
  end;
  if ToWidth < 1 then ToWidth := 1;
  if ToHeight < 1 then ToHeight := 1;

  cX := 1/2 * ToWidth;
  cY := 1/2 * ToHeight;
  cDiag := SqRt(cX*cX + cY*cY);
  ResizeGLScene(ToWidth, ToHeight, cwX, cwY);
  glEnable(GL_TEXTURE_2D);
  SetAntialias(True);
  glEnable(GL_BLEND);
  //glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); //Alpha
  glBlendFunc(GL_SRC_ALPHA, GL_ONE); //Additive

  SetLength(TexWord, TexSz*TexSz);
  for iTexel := 0 to High(TexWord) do TexWord[iTexel] := $00FF;
  {
  TexWord[cLo + TexSz * cLo] := $FFFF;
  TexWord[cHi + TexSz * cLo] := $FFFF;
  TexWord[cLo + TexSz * cHi] := $FFFF;
  TexWord[cHi + TexSz * cHi] := $FFFF;
  }
  for iY := 0 to TexSz-1 do
    for iX := (TexSz div 2)-4 to (TexSz div 2)+3 do
      TexWord[iX + TexSz * iY] := $FFFF
  ;
  hStarTex := AllocateTex;
  SetBoundTexture(hStarTex);
  //glTexImage2D(GL_TEXTURE_2D, 0, GL_LUMINANCE_ALPHA, TexSz, TexSz, 0, GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, @TexWord[0]); //`@ Noworky. Why?
  gluBuild2DMipmaps(GL_TEXTURE_2D, GL_LUMINANCE_ALPHA, TexSz, TexSz, GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, @TexWord[0]); //I think we want mipmaps anyways, I just can't test.

  Err := glGetError;
  if Err <> GL_NO_ERROR then
    AnsiMessageBox('Error ' + IntToStr(Err), 'SuperStarSaver 2', MB_ICONEXCLAMATION)
  ;

  SetLength(StarData, StarCount);
  for iStar := 0 to High(StarData) do StarData[iStar] := TStarData.Create;
  bNeedsSort := True;

  Application.OnIdle := Aperture.IdleFrame;
  MasterClock.ExtractIDT;
end;

procedure StopScreensaver;
var
  iStar: Integer;

begin
  Application.OnIdle := nil;

  FreeAllocatedTextures;

  FreeAndNil(Aperture);
  FreeAndNil(MasterClock);

  for iStar := 0 to High(StarData) do try
    StarData[iStar].Free;
  except
  end;
end;

{ TOpenGLApertureWithFrame }

procedure TOpenGLApertureWithFrame.IdleFrame(Sender: TObject; var Done: Boolean);
var
  dT, dExp, dZ: TReal;
  iStar: Integer;

begin
  try
    Done := False;

    dT := MasterClock.ExtractDT;

    dZ := -SpeedFactor * dT;
    dExp := Power(1/2, -1/1024*dZ);
    for iStar := 0 to High(StarData) do
      StarData[iStar].DoMove(dExp, dZ)
    ;

    glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);

    for iStar := 0 to High(StarData) do
      StarData[iStar].DoDraw
    ;

    Aperture.FinishFrame;

  except
    Application.OnIdle := nil;
    Done := True;
    raise;
  end;
end;

{ TConfigForm }

procedure TConfigForm.CancelButtonClick(Sender: TObject);
begin
  Close;
end;

procedure TConfigForm.FormCreate(Sender: TObject);
begin
  LoadDefaultFont(Font);

  StarHead := StarLabel.Caption;
  SpeedHead := SpeedLabel.Caption;

  bProgrammaticChange := True;
  StarTrack.Position := Round(Log2(StarCount) * 10);
  SpeedTrack.Position := Round(Log2(SpeedFactor) * 10);
  bProgrammaticChange := False;

  StarTrackChange(StarTrack);
  SpeedTrackChange(SpeedTrack);
end;

procedure TConfigForm.FormDestroy(Sender: TObject);
begin
  //`@
end;

procedure TConfigForm.OKButtonClick(Sender: TObject);
var
  Reg: TRegistry;
  bWroteOK: Boolean;

begin
  bWroteOK := True;

  Reg := TRegistry.Create(KEY_CREATE_SUB_KEY or KEY_SET_VALUE);
  Reg.RootKey := HKEY_CURRENT_USER;
  if not Reg.OpenKey(MyKey, True) then begin
    bWroteOK := False;
  end
  else begin
    try
      Reg.WriteString('SpeedFactor', FloatToStr(SpeedFactor));
    except
      bWroteOK := False;
    end;
    try
      Reg.WriteInteger('StarCount', StarCount);
    except
      bWroteOK := False;
    end;
  end;
  Reg.Free;

  if bWroteOK then Close;
end;

procedure TConfigForm.SpeedTrackChange(Sender: TObject);
var
  WarpFactor: TReal;

begin
  if bProgrammaticChange then Exit;

  SpeedFactor := Power(2, SpeedTrack.Position/10);
  WarpFactor := Power(1/32 * SpeedFactor, 1/3);
  WarpFactor := Round(WarpFactor * 20) / 20;
  SpeedLabel.Caption := SpeedHead + FloatToStr(WarpFactor);
end;

procedure TConfigForm.StarTrackChange(Sender: TObject);
begin
  if bProgrammaticChange then Exit;

  StarCount := Round(Power(2, StarTrack.Position/10));
  StarLabel.Caption := StarHead + IntToStr(StarCount);
end;

{ TStarData }

procedure TStarData.BeNewStar(bAnywhere: Boolean);
const
  StarWide = 4;

var
  Colorness, Hyp: TReal;

begin
  Loc.Z := FarClip;
  with Loc do repeat
    X := FarClip * RandBipolar;
    Y := FarClip * RandBipolar;
    if bAnywhere then Z := FarClip * RandReal;
  until ((Abs(X) >= 8) or (Abs(Y) >= 8)) and bInFrustum;

  with Loc do begin
    Hyp := SqRt(X*X + Y*Y);
    SideStep := Vector(+StarWide*Y/Hyp, -StarWide*X/Hyp, 0);

    Hyp := (Hyp - 1)/Hyp;
    oLoc := Vector(X*Hyp, Y*Hyp, Z);
  end;

  Colorness := 1/8 * RandReal;
  Color.G := 1 - Colorness;
  if RandBit then begin
    Color.R := 1;
    Color.B := Color.G;
  end
  else begin
    Color.B := 1;
    Color.R := Color.G;
  end;
end;

function TStarData.bInFrustum: Boolean;
begin
  with Loc do Result := (Z >= NearClip) and (Abs(X)/Z < cwX) and (Abs(Y)/Z < cwY);
end;

constructor TStarData.Create;
begin
  //inherited;

  BeNewStar(True);
end;

procedure TStarData.DoDraw;
begin
  with Color do SetColor(R, G, B);

  glBegin(GL_TRIANGLES);
    glTexCoord2d(1/2, 1/2); with oLoc do glVertex3d(X, Y, Z);
    glTexCoord2d(0  , 1/2); with VSum(Loc, SideStep) do glVertex3d(X, Y, Z);
    glTexCoord2d(1  , 1/2); with VDiff(Loc, SideStep) do glVertex3d(X, Y, Z);
  glEnd;
end;

procedure TStarData.DoMove(M, B: TReal);
begin
  if not bInFrustum then 
    BeNewStar(False)
  ;

  oLoc := Loc;
  Loc.Z := Loc.Z * M + B;
end;

end.
